/*
 *
 * Copyright 2015, Google Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "grpc/_adapter/_tag.h"

#include <Python.h>
#include <grpc/grpc.h>
#include <grpc/support/alloc.h>

pygrpc_tag *pygrpc_tag_new(pygrpc_tag_type type, PyObject *user_tag,
                           Call *call) {
  pygrpc_tag *self = (pygrpc_tag *)gpr_malloc(sizeof(pygrpc_tag));
  memset(self, 0, sizeof(pygrpc_tag));
  if (user_tag == NULL) {
    self->user_tag = Py_None;
  } else {
    self->user_tag = user_tag;
  }
  Py_INCREF(self->user_tag);
  self->type = type;
  self->call = call;
  Py_INCREF(call);
  return self;
}

pygrpc_tag *pygrpc_tag_new_server_rpc_call(PyObject *user_tag) {
  return pygrpc_tag_new(PYGRPC_SERVER_RPC_NEW, user_tag,
                        (Call *)pygrpc_CallType.tp_alloc(&pygrpc_CallType, 0));
}

void pygrpc_tag_destroy(pygrpc_tag *self) {
  Py_XDECREF(self->user_tag);
  Py_XDECREF(self->call);
  gpr_free(self);
}
